/* $Log: RemoteToolkit.java,v $
 * Revision 1.1  1998/04/30  19:46:29  stuart
 * Initial revision
 *
 */
package bmsi.tui;

import bmsi.awt.TextLog;
import bmsi.xopen.FType;
import java.awt.*;
import java.awt.event.*;
import java.awt.peer.*;
import java.awt.image.*;
import java.awt.datatransfer.Clipboard;
import java.util.Properties;
import java.util.Vector;
import java.io.*;
import java.net.*;

/** RemoteToolkit is an AWT Toolkit implementation for text terminals.
 * TUIKit communicates with a "client" program via a socket.  You could
 * also consider the "client" a "display server", but the "client" initiates
 * the connection.  
 * After construction, the conversation consists of method calls.  A method
 * call is an object id, a method index, optional parameter values, and a
 * possible return value.  Value types are object id, short, or string.
 */

public abstract class RemoteToolkit extends Toolkit
  implements Runnable, RemotePeer
{
  private TextLog log;
  private Properties props;
  /** Used by TUIComponents for this toolkit to track FocusEvents */
  TUIComponent hasFocus;
  TUIWindow active;
  EventQueue theQueue;
  private boolean needFlush = false;
  private boolean singleUser = true;
  private TUISync flushTarget = new TUISync(this);
  private Socket sock = null;
  private DataOutputStream out;
  private DataInputStream in;
  private Dimension screen;
  private Dimension scale = new Dimension(1,1);
  private static ServerSocket server;
  private static int serverPort = 9713;
  private Thread eventThread = new Thread(this,"TUIKitEvents");
  private Vector objs = new Vector();
  private Object retval = null;
  Vector fontvect;
  private String[] fontlist = { "Terminal" };
  private byte[] edxParm = null;

  /** Strings are converted to an 8-bit form easily understood by the remote
      peer.  Our TUI system uses the IBM PC codepage. */
  private FType codepage;

  /** If setPort() is called with a non-zero port, this Toolkit waits for a
    connection from the "client" on that port.  Otherwise, this Toolkit
    assumes that this JVM was loaded via inetd and uses FileDescriptor.out
   */
  public static void setPort(int port) {
    if (server != null) {
      try { server.close(); }
      catch (IOException e) { }
      server = null;
    }
    serverPort = port;
  }

  /** Select the encoding used to exchange Strings with our remote peer.
      The default encoding is Cp437 (IBMPC) for TUIKit and UTF8 for
      GUIKit.
   */
  public void setEncoding(String cpname) throws IOException {
    codepage = FType.getConverter(cpname);
  }

  /** Enable use of an extension package.  We don't support any extensions
      yet. */
  public void setExtension(String extname) {
  }

  /** Specify the class to notified of properties sent from the remote
      client.  This typically includes things like <code>user.dir</code> and
      <code>user.name</code> and application specific environment settings
      which are useful to a multiuser 
      login package for Java.  (To allow multiple remote clients to
      share the same JVM.)  This property can be set only once.
   */
  public synchronized void setPropertyListener(Properties p) {
    if (props != null)
      throw new SecurityException(
	"PropertyListener already set for RemoteToolkit");
    props = p;
  }

  public RemoteToolkit() throws IOException {
    Properties prop = System.getProperties();
    String tuipeer = prop.getProperty("bmsi.tuipeer");
    if (tuipeer == null && serverPort > 0 && server == null) {
      server = new ServerSocket(serverPort);
    }
    if (server != null) {
      debug("Waiting for connection...");
      sock = server.accept();
      debug("Accepting call from " + sock);
    }
    else {
      // check whether client autoload enabled
      if (tuipeer != null) {
	ServerSocket server = new ServerSocket(0);
	String cmd = tuipeer + " -p" + server.getLocalPort();
	Process prog = Runtime.getRuntime().exec(cmd);
	sock = server.accept();
	server.close();
	// check that client is on localhost
	InetAddress them = sock.getInetAddress();
	String host = them.getHostName();
	String ourhost = InetAddress.getLocalHost().getHostName();
	if (!them.equals(InetAddress.getLocalHost()) && !host.equals(ourhost)
	   && !host.equals("localhost")) {
	  sock.close();
	  throw new AWTError("Autoloaded AWT client not on "
		+ ourhost + ": " + host);
	}
        System.err.println("Connection from " + them + " to " + ourhost);
	sock.setTcpNoDelay(true);
	//System.runFinalizersOnExit(true);
      }
      //log = new TextLog(20,60);
    }
    if (sock != null)
      init(sock.getInputStream(),sock.getOutputStream());
    else
      init(System.in,System.out);
    eventThread.start();
  }

  public RemoteToolkit(Socket sock) throws IOException {
    this.sock = sock;
    singleUser = false;
    init(sock.getInputStream(),sock.getOutputStream());
    eventThread.start();
  }

  private void init(InputStream i,OutputStream o) throws IOException {
    setEncoding("Cp437");
    o = new BufferedOutputStream(o);
    out = new DataOutputStream(o);
    in = new DataInputStream(i);
    objs.addElement(this);
    theQueue = new EventQueue();
    while (screen == null)
      readCmd();	// execute commands until Screen size is defined
  }

  public void run() {
    if (log != null) {
      Frame f = new Frame("Java Error Log");
      f.add(log);
      f.pack();
      f.show();
      System.setErr(log.getPrintStream());
      System.setOut(log.getPrintStream());
    }
    debug("Screen = " + screen);
    debug(" Scale = " + scale);
    try {
      for (;;)
	readCmd();
    }
    // remote peer has gone away
    catch (AWTError e) {
      debug(e);
      // exit if single user AWT
      if (singleUser)
	System.exit(0);
    }
  }

  public void debug(String ln) { }

  public final void debug(Throwable e) {
    e.printStackTrace();
  }

  /** Scale a TUI dimension to defeat hardwired insets and sizes that
      are ridiculously large for a TUI interface. 
    @parm dim	The Dimension to scale in text positions.  
		The argument is modified in place.
    @return 	The Dimension scaled to simulated pixels.
   */
  final Dimension scale(Dimension dim) {
    dim.width *= scale.width;
    dim.height *= scale.height;
    return dim;
  }

  /** wait for an input event from our remote peer. */
  private void readCmd() {
    try {
      //flush();
      int id = in.readShort();
      int cmd = in.readShort();
      debug("readCmd: " + objs.elementAt(id) + "(" + id + "), " + cmd);
      RemotePeer peer = (RemotePeer)objs.elementAt(id);
      peer.remoteMethod(cmd);
    }
    catch (IOException e) {
      throw new AWTError(e.toString());
    }
  }

  public Object getTarget() { return null; }
  public int getID() { return 0; }

  /** Toolkit remote methods.  This Toolkit has a well known ID (0)
   * and supports some special methods.
   */
  public void remoteMethod(int cmd) throws IOException {
    int w,h;
    switch (cmd) {
    case SETEXTENSION:
      setExtension(readUTF());
      return;
    case SETTEXT:	// select String encoding
      setEncoding(readUTF());
      return;
    case SETFONT:	// add to Fontlist
      if (fontvect == null)
	fontvect = new Vector();
      fontvect.addElement(readUTF());
      return;
    case RETSHORT:	// return a 2 byte integer
      retObject(new Integer(in.readShort()));
      return;
    case RETSTRING:
      retObject(readUTF());
      return;
    case RETDIM:
      w = in.readShort();
      h = in.readShort();
      retObject(new Dimension(w,h));
      return;
    case RETINSETS:
      retObject(
	new Insets(in.readShort(),in.readShort(),in.readShort(),in.readShort())
      );
      return;
    case RETSHORTARRAY: {
      int len = in.readUnsignedShort();
      char[] a = new char[len];
      for (int i = 0; i < len; ++i)
	a[i] = (char)in.readUnsignedShort();
      retObject(a);
      return;
    }
    case SETPARM: {
      int len = in.readUnsignedShort();
      edxParm = new byte[len];
      in.readFully(edxParm);
      return;
    }
    case SETPROP: {
      String key = readUTF();
      String val = readUTF();
      try {
	if (props != null)
	  props.put(key,val);
	else
	  System.getProperties().put(key,val);
      }
      catch (Exception e) { }	// don't let exceptions kill us
      return;
    }
    case RETFONT: {
      String name = readUTF();
      int size = in.readShort();
      int style = in.readShort();
      Font f = new Font(name,style,size);
      retObject(f);
      return;
    }
    case SETSCALE:
      scale.width = in.readShort();
      scale.height = in.readShort();
      return;
    case SCREENSIZE:
      w = in.readShort();
      h = in.readShort();
      screen = new Dimension(w,h);
      return;
    case RUNCLASS:
      Viewer.runClass(readUTF(),this);
      return;
    }
  }

  final void writeShort(int val) {
    try {
      out.writeShort(val);
    }
    catch (IOException e) {
      throw new AWTError(e.toString());
    }
  }

  /** Send an encoded String to our remote peer.  The name is historical:
    UTF8 was originally the only encoding supported. */
  private void writeUTF(String s) {
    try {
      byte[] buf = codepage.getBytes(s);
      if (buf.length > 0xFFFF)
	throw new AWTError("String too long");
      out.writeShort(buf.length);
      out.write(buf);
    }
    catch (IOException e) {
      throw new AWTError(e.toString());
    }
  }

  /** Read an encoded String from our remote peer.  The name is historical:
    UTF8 was originally the only encoding supported. */
  synchronized String readUTF() throws IOException {
    byte[] buf = new byte[readShort() & 0xFFFF];
    in.readFully(buf);
    return codepage.getString(buf);
  }

  synchronized int readShort() throws IOException { return in.readShort(); }

  final int readUnsignedShort() throws IOException {
    return readShort() & 0xFFFF;
  }

  /** get the next returned integer value */
  final int getIntegerRet() {
    return ((Number)getObjectRet()).intValue();
  }

  /** get the next return String value */
  final String getStringRet() {
    return getObjectRet().toString();
  }

  /** get the next return value as a Dimension */
  final Dimension getDimRet() {
    return (Dimension)getObjectRet();
  }

  /** get the new return value as a char[] */
  final char[] getCharArrayRet() {
    return (char[])getObjectRet();
  }

  final Insets getInsetRet() {
    return (Insets)getObjectRet();
  }

  /** get the next Object returned from our remote peer. */
  synchronized Object getObjectRet() {
    try {
      while (retval == null) {
	if (Thread.currentThread() == eventThread)
	  readCmd();
	else {
	  flush();
	  wait();
	}
      }
      Object val = retval;
      retval = null;
      notify();
      return val;
    }
    catch (InterruptedException e) {
      throw new AWTError("Interrupted waiting for a return value");
    }
  }

  /** post an object returned by our remote peer. */
  synchronized void retObject(Object val) throws IOException {
    try {
      while (retval != null)
	wait();
      retval = val;
      notify();
    }
    catch (InterruptedException e) {
      throw new IOException("Interrupted waiting to return a value");
    }
  }

  /** Send a simple command to our remote peer. */
  synchronized void writeCmd(RemotePeer peer,int cmd) {
    writeCmd(peer.getID(),cmd);
    if (!needFlush) {
      needFlush = true;
      flushTarget.queue();
    }
  }

  private synchronized void writeCmd(int id,int cmd) {
    //debug("CMD: " + objs.elementAt(id) + "(" + id + "), " + cmd);
    writeShort(id);
    writeShort(cmd);
  }

  /** Flush socket output to our remote peer.  */
  public synchronized void flush() {
    try {
      out.flush();
      needFlush = false;
      //debug("Flush");
    }
    catch (IOException e) {
      throw new AWTError(e.toString());
    }
  }

  // various remote commands with arguments follow -

  synchronized void writeCmd(RemotePeer id,int cmd,int val) {
    writeCmd(id,cmd);
    writeShort(val);
  }

  synchronized void writeCmd(RemotePeer id,int cmd,String s) {
    writeCmd(id,cmd);
    writeUTF(s);
  }

  synchronized void writeCmd(RemotePeer id,int cmd,int x,int y) {
    writeCmd(id,cmd);
    writeShort(x);
    writeShort(y);
  }

  synchronized void writeCmd(RemotePeer id,int cmd,String s,int x) {
    writeCmd(id,cmd);
    writeUTF(s);
    writeShort(x);
  }

  synchronized void writeCmd(RemotePeer id,int cmd,String s,int x,int y) {
    writeCmd(id,cmd);
    writeUTF(s);
    writeShort(x);
    writeShort(y);
  }

  synchronized void writeCmd(RemotePeer id,int cmd,int x,int y,int w,int h,
	int dx,int dy)
  {
    writeCmd(id,cmd);
    writeShort(x);
    writeShort(y);
    writeShort(w);
    writeShort(h);
    writeShort(dx);
    writeShort(dy);
  }

  synchronized void writeCmd(RemotePeer id,int cmd,int x,int y,int w,int h) {
    writeCmd(id,cmd);
    writeShort(x);
    writeShort(y);
    writeShort(w);
    writeShort(h);
  }

  synchronized void writeCmd(RemotePeer id,int cmd,RemotePeer c,int x,int y) {
    writeCmd(id,cmd);
    writeShort(c.getID());
    writeShort(x);
    writeShort(y);
  }

  public final byte[] getEDXParm() {
    return edxParm;
  }

  public synchronized void close() throws IOException {
    try {
      out.flush();
    }
    finally {
      if (sock != null)
	sock.close();
    }
  }

  /** Make sure the remote connection always gets closed.
     @exception Throwable	If we can't close the remote connection.
   */

  protected void finalize() throws IOException {
    close();
  }


  /** Create a generic remote peer.  We generate and keep track of the remote
   object ids locally within Java so that we don't have to wait for the remote
   program to respond with a new id.  Turnaround on a remote socket is
   a performance killer.
   */
  synchronized int createRemotePeer(RemotePeer par,int type,RemotePeer peer) {
    if (par == null)
      par = this;
    /* Vector won't search for null, so we search for ourselves knowing
       that a real TUIKit reference will be in pos 0 only */
    int id = objs.indexOf(this,1);
    if (id < 0) {
      id = objs.size();
      objs.addElement(peer);
    }
    else
      objs.setElementAt(peer,id);
    writeCmd(par,type,id);
    return id;
  }

  /** Tell the remote program to remove an object. */
  synchronized void removePeer(int id) {
    if (id < 0) return;
    objs.setElementAt(this,id);
    writeCmd(id,DISPOSE);	// tell the remote peer to go away
    flush();
  }

  public Dimension getScreenSize() { return screen; }
  public int getScreenResolution() { return 10; } // "pixels"/inch
  public synchronized String[] getFontList() {
    if (fontvect != null) {
      fontlist = new String[fontvect.size()];
      fontvect.copyInto(fontlist);
      fontvect = null;
    }
    return fontlist;
  }

  public void sync() { writeCmd(0,SYNC); flush(); }

  // TUI's can't reasonably support Image.  Always fail -
  //  However, we might want to support in memory text arrays as
  //  a Canvas with no parent.

  public Image getImage(String name) { return null; }
  public Image getImage(java.net.URL url) { return null; }
  public boolean prepareImage(Image img,int w,int h,ImageObserver obs) {
    return false;
  }
  public int checkImage(Image img,int w,int h,ImageObserver obs) {
    return ImageObserver.ERROR | ImageObserver.ABORT;
  }
  public Image createImage(ImageProducer p) { return null; }
  public Image createImage(byte[] data,int offset,int len) { return null; }

  // JDK 1.1 features -

  protected EventQueue getSystemEventQueueImpl() { return theQueue; }

  /** Anybody can beep! */
  public void beep() { writeCmd(0,BEEP); flush(); }

  /** This JDK1.1 feature has no documentation! This seems to be a placeholder
      that is not yet used for anything.
   */
  protected FontPeer getFontPeer(String name, int style) {
    //debug("FontPeer needed");
    return null;
  }

  /** TUI's have a clipboard, but we haven't finished this JDK1.1 feature. */
  public Clipboard getSystemClipboard() {
    debug("ClipBoard needed");
    return null;
  }

  /** TUI's have printers, but we haven't finished this JDK1.1 feature. */
  public PrintJob getPrintJob(Frame frame, String jobtitle, Properties props) {
    debug("PrintJob needed");
    return null;
  }

  /** TUI's have ScrollPanes, but we haven't finished this JDK1.1 feature. */
  protected ScrollPanePeer createScrollPane(ScrollPane target) {
    debug("ScrollPane needed");
    return null;
  }

  /** Create a TUIKit on Filedescriptor.out and make it the default Toolkit.  
    Used when loading a Java VM from inetd. */
  public static void main(String[] args) {
    Properties props = System.getProperties();
    props.put("awt.toolkit","bmsi.tui.TUIKit");
    setPort(0);
    Toolkit.getDefaultToolkit();
  }

  // we have to redeclare all these so they can be accessed by 
  // MultiuserToolkit

  protected abstract ButtonPeer createButton(Button b);
  protected abstract TextFieldPeer createTextField(TextField t);
  protected abstract LabelPeer createLabel(Label lbl);
  protected abstract ListPeer createList(List lst);
  protected abstract TextAreaPeer createTextArea(TextArea t);
  public abstract FontMetrics getFontMetrics(Font f);
  public abstract ColorModel getColorModel();
  protected abstract CheckboxPeer createCheckbox(Checkbox cb);
  protected abstract ScrollbarPeer createScrollbar(Scrollbar s);
  protected abstract ChoicePeer createChoice(Choice c);
  protected abstract FramePeer createFrame(Frame f);
  protected abstract CanvasPeer createCanvas(Canvas c);
  protected abstract PanelPeer createPanel(Panel p);
  protected abstract WindowPeer createWindow(Window w);
  protected abstract DialogPeer createDialog(Dialog dlg);
  protected abstract MenuBarPeer createMenuBar(MenuBar mnu);
  protected abstract MenuPeer createMenu(Menu mnu);
  protected abstract MenuItemPeer createMenuItem(MenuItem mi);
  protected abstract FileDialogPeer createFileDialog(FileDialog dlg);
  protected abstract CheckboxMenuItemPeer
    createCheckboxMenuItem(CheckboxMenuItem mi);
  protected abstract PopupMenuPeer createPopupMenu(PopupMenu target);

  /** A BMS extension intended to display documentation on the
     remote peer.
   */

  public void showDocument(URL url) {
    writeCmd(this,SHOWURL,url.toString());
  }
}
